# Legacy module of petisco version 0

from petisco.legacy.logger.interface_logger import ILogger
from petisco.legacy.logger.log_message import LogMessage
from petisco.legacy.logger.logging_based_logger import LoggingBasedLogger

classes = ["ILogger", "LogMessage", "LoggingBasedLogger"]

# Constants
from petisco.legacy.logger.interface_logger import (
    CRITICAL,
    DEBUG,
    ERROR,
    FATAL,
    INFO,
    NOTSET,
    WARN,
    WARNING,
)

constants = ["CRITICAL", "FATAL", "ERROR", "WARNING", "WARN", "INFO", "DEBUG", "NOTSET"]

__all__ = classes + constants



from typing import Any, Dict

from meiga import AnyResult

from petisco.extra.fastapi.controller.fastapi_default_response import (
    FASTAPI_DEFAULT_RESPONSE,
)


def fastapi_success_handler(result: AnyResult) -> Dict[str, Any]:
    try:
        response = (
            FASTAPI_DEFAULT_RESPONSE
            if result.value is True
            else {"result": result.value}
        )
    except Exception:  # noqa
        response = FASTAPI_DEFAULT_RESPONSE

    return response



import os
import threading
import traceback
from dataclasses import dataclass
from time import sleep
from typing import Any, Callable, Dict, List, NoReturn, Optional, Type, Union

from loguru import logger
from meiga import Failure
from pika import BasicProperties
from pika.adapters.blocking_connection import BlockingChannel
from pika.exceptions import ConnectionClosedByBroker
from pika.spec import Basic

from petisco.base.domain.message.chaos.message_chaos import MessageChaos
from petisco.base.domain.message.chaos.message_chaos_error import MessageChaosError
from petisco.base.domain.message.chaos.not_implemented_message_chaos import (
    NotImplementedMessageChaos,
)
from petisco.base.domain.message.command import Command
from petisco.base.domain.message.consumer_derived_action import ConsumerDerivedAction
from petisco.base.domain.message.domain_event import DomainEvent
from petisco.base.domain.message.message import Message
from petisco.base.domain.message.message_consumer import MessageConsumer
from petisco.base.domain.message.message_handler_returns_none_error import (
    MessageHandlerReturnsNoneError,
)
from petisco.base.domain.message.message_subscriber import MessageSubscriber
from petisco.extra.rabbitmq.application.message.bus.rabbitmq_command_bus import (
    RabbitMqCommandBus,
)
from petisco.extra.rabbitmq.application.message.bus.rabbitmq_domain_event_bus import (
    RabbitMqDomainEventBus,
)
from petisco.extra.rabbitmq.application.message.consumer.rabbitmq_consumer_connector import (
    RabbitMqConsumerConnector,
)
from petisco.extra.rabbitmq.application.message.consumer.rabbitmq_event_consumer_logger import (
    RabbitMqMessageConsumerLogger,
)
from petisco.extra.rabbitmq.application.message.consumer.rabbitmq_event_consumer_printer import (
    RabbitMqEventConsumerPrinter,
)
from petisco.extra.rabbitmq.application.message.formatter.rabbitmq_message_subscriber_queue_name_formatter import (
    RabbitMqMessageSubscriberQueueNameFormatter,
)
from petisco.extra.rabbitmq.shared.rabbitmq_connector import RabbitMqConnector
from petisco.extra.rabbitmq.shared.rabbitmq_exchange_name_formatter import (
    RabbitMqExchangeNameFormatter,
)
from petisco.legacy.logger.interface_logger import ILogger
from petisco.legacy.logger.not_implemented_logger import NotImplementedLogger

MAX_ATTEMPTS_TO_RECONNECT = int(
    os.getenv("PETISCO_RABBITMQ_MAX_ATTEMPTS_TO_RECONNECT_CONSUMER", "20")
)
WAIT_SECONDS_TO_RECONNECT = int(
    os.getenv("PETISCO_RABBITMQ_WAIT_SECONDS_TO_RECONNECT_CONSUMER", "5")
)


@dataclass
class SubscriberItem:
    queue_name: str
    subscriber: MessageSubscriber
    consumer_tag: str
    is_store: bool = False


class RabbitMqMessageConsumer(MessageConsumer):
    """
    A RabbitMQ consumer to link received messages from rabbitmq with defined subscribers.
    """

    def __init__(
        self,
        organization: str,
        service: str,
        max_retries: int,
        connector: RabbitMqConnector = RabbitMqConnector(),
        verbose: bool = False,
        chaos: MessageChaos = NotImplementedMessageChaos(),
        logger: Optional[ILogger] = NotImplementedLogger(),
    ) -> None:
        self.connector = connector
        self.organization = organization
        self.service = service
        self.exchange_name = f"{organization}.{service}"
        self.rabbitmq_key = f"consumer-{self.exchange_name}"
        self._fallback_store_exchange_name = f"retry.{organization}.store"
        self.max_retries = max_retries
        self._channel = self.connector.get_channel(self.rabbitmq_key)
        self.printer = RabbitMqEventConsumerPrinter(verbose)
        self.consumer_logger = RabbitMqMessageConsumerLogger(logger)
        self.chaos = chaos
        self.subscribers: Dict[str, SubscriberItem] = {}
        self._thread: Union[threading.Thread, None] = None

    def start(self) -> NoReturn:
        """
        Start the process of consuming messages from RabbitMQ and pass to subscriber.
        """
        if not self._channel:
            raise RuntimeError(
                "RabbitMqMessageConsumer: cannot start consuming event without any subscriber defined."
            )
        self._thread = threading.Thread(target=self._start)
        self._thread.start()

    def _start(self) -> None:
        try:
            self._channel.start_consuming()
        except ConnectionClosedByBroker:
            self._re_connect(attempt=1)

    def _re_connect(self, attempt: int):
        if attempt >= MAX_ATTEMPTS_TO_RECONNECT:
            raise ConnectionError(
                f"Impossible to reconnect consumer '{self.rabbitmq_key}' after {attempt} attempts"
            )

        logger.info(
            f"Trying to reconnect consumer '{self.rabbitmq_key}' (attempt {attempt})"
        )

        try:
            self._channel = self.connector.get_channel(self.rabbitmq_key)
        except ConnectionError:
            sleep(WAIT_SECONDS_TO_RECONNECT)
            attempt += 1
            self._re_connect(attempt=attempt)
        else:
            subscribers = [
                item.subscriber.__class__ for item in self.subscribers.values()
            ]
            self.add_subscribers(subscribers)
            self.start()
            logger.info(
                f"Consumer '{self.rabbitmq_key}' reconnected after {attempt} attempts"
            )

    def add_subscribers(self, subscribers: List[Type[MessageSubscriber]]) -> None:
        """
        Add defined subscribers to be connected with main queues.
        """

        for SubscriberClass in subscribers:

            subscriber: MessageSubscriber = SubscriberClass()

            for subscriber_info in subscriber.get_message_subscribers_info():

                if subscriber_info.message_type == "message":
                    self._add_subscriber_instance_on_queue(
                        queue_name="store", subscriber=subscriber, is_store=True
                    )
                else:
                    queue_name = RabbitMqMessageSubscriberQueueNameFormatter.format(
                        subscriber_info, exchange_name=self.exchange_name
                    )
                    self._add_subscriber_instance_on_queue(
                        queue_name=f"{queue_name}.{subscriber.get_subscriber_name()}",
                        subscriber=subscriber,
                        message_type_expected=subscriber_info.message_type,
                    )

    def add_subscriber_on_dead_letter(
        self, subscriber: Type[MessageSubscriber]
    ) -> None:
        """
        Add defined subscribers to be connected with the correspondent dead letter.
        """
        instance_subscriber: MessageSubscriber = subscriber()
        for subscriber_info in instance_subscriber.get_message_subscribers_info():

            queue_name = RabbitMqMessageSubscriberQueueNameFormatter.format_dead_letter(
                subscriber_info, exchange_name=self.exchange_name
            )
            self._add_subscriber_instance_on_queue(
                queue_name=f"{queue_name}.{instance_subscriber.get_subscriber_name()}",
                subscriber=instance_subscriber,
                message_type_expected=subscriber_info.message_type,
            )

    def _add_subscriber_instance_on_queue(
        self,
        queue_name: str,
        subscriber: MessageSubscriber,
        is_store: bool = False,
        message_type_expected: str = "message",
    ) -> None:
        consumer_tag = self._channel.basic_consume(
            queue=queue_name,
            on_message_callback=self.consumer(
                subscriber, is_store, message_type_expected
            ),
        )
        self.subscribers[queue_name] = SubscriberItem(
            queue_name, subscriber, consumer_tag, is_store
        )

    def add_subscriber_on_queue(
        self,
        queue_name: str,
        subscriber: Type[MessageSubscriber],
        is_store: bool = False,
        message_type_expected: str = "message",
    ) -> None:
        """
        Add defined subscribers to be connected with a specific queue name.
        """
        subscriber_instance: MessageSubscriber = subscriber()
        self._add_subscriber_instance_on_queue(
            queue_name=queue_name,
            subscriber=subscriber_instance,
            is_store=is_store,
            message_type_expected=message_type_expected,
        )

    def consumer(
        self,
        subscriber: MessageSubscriber,
        is_store: bool = False,
        message_type_expected: str = "message",
    ) -> Callable[..., None]:
        def rabbitmq_consumer(
            ch: BlockingChannel,
            method: Basic.Deliver,
            properties: BasicProperties,
            body: bytes,
        ) -> None:
            self.printer.print_received_message(method, properties, body)

            if self.chaos.nack_simulation(ch, method):
                self.consumer_logger.log_nack_simulation(
                    method, properties, body, subscriber.handle
                )
                return
            else:
                self.consumer_logger.log(
                    method,
                    properties,
                    body,
                    subscriber.handle,
                    log_activity="received_message",
                )
            try:
                if message_type_expected == "domain_event":
                    message = DomainEvent.from_json(body)
                elif message_type_expected == "command":
                    message = Command.from_json(body)
                else:
                    message = Message.from_json(body)
            except Exception as e:
                self.consumer_logger.log_parser_error(
                    method, properties, body, subscriber.handle, e
                )
                ch.basic_nack(delivery_tag=method.delivery_tag)
                return

            self.chaos.delay()

            if self.chaos.failure_simulation(method):
                self.consumer_logger.log_failure_simulation(
                    method, properties, body, subscriber.handle
                )
                result = Failure(MessageChaosError())
            else:
                connector = RabbitMqConsumerConnector(ch)
                domain_event_bus = RabbitMqDomainEventBus(
                    self.organization, self.service, connector
                )
                command_bus = RabbitMqCommandBus(
                    self.organization, self.service, connector
                )
                subscriber.set_domain_event_bus(domain_event_bus)
                subscriber.set_command_bus(command_bus)
                result = subscriber.handle(message)

            self.printer.print_context(subscriber.handle, result)

            if result is None:
                raise MessageHandlerReturnsNoneError(subscriber.handle)

            derived_action = ConsumerDerivedAction()

            if result.is_failure:
                if not properties.headers:
                    properties.headers = {
                        "queue": f"{method.routing_key}.{subscriber.get_subscriber_name()}"
                    }
                derived_action = self.handle_consumption_error(
                    ch, method, properties, body, is_store
                )
            else:
                ch.basic_ack(delivery_tag=method.delivery_tag)

            self.consumer_logger.log(
                method,
                properties,
                body,
                subscriber.handle,
                "computed_message",
                result,
                derived_action,
            )
            self.printer.print_separator()

        return rabbitmq_consumer

    def handle_consumption_error(
        self,
        ch: BlockingChannel,
        method: Basic.Deliver,
        properties: BasicProperties,
        body: bytes,
        is_store: bool,
    ) -> ConsumerDerivedAction:

        if self.has_been_redelivered_too_much(properties):
            derived_action = self.send_to_dead_letter(ch, method, properties, body)
        else:
            derived_action = self.send_to_retry(ch, method, properties, body, is_store)

        ch.basic_ack(delivery_tag=method.delivery_tag)

        return derived_action

    def has_been_redelivered_too_much(self, properties: BasicProperties) -> bool:
        if not properties.headers or "redelivery_count" not in properties.headers:
            if self.max_retries < 1:
                return True
            return False
        else:
            redelivery_count: int = properties.headers.get("redelivery_count")
            return bool(redelivery_count >= self.max_retries)

    def _get_routing_key(self, routing_key: str, prefix: str) -> str:
        if routing_key.startswith("retry."):
            routing_key = routing_key.replace("retry.", prefix, 1)
        elif routing_key.startswith("dead_letter."):
            routing_key = routing_key.replace("dead_letter.", prefix, 1)
        else:
            routing_key = f"{prefix}{routing_key}"
        return routing_key

    def send_to_retry(
        self,
        ch: BlockingChannel,
        method: Basic.Deliver,
        properties: BasicProperties,
        body: bytes,
        is_store: bool = False,
    ) -> ConsumerDerivedAction:
        self.printer.print_action("send_to_retry")
        exchange_name = RabbitMqExchangeNameFormatter.retry(self.exchange_name)

        routing_key = method.routing_key
        if properties.headers:
            routing_key = properties.headers.get("queue", routing_key)

        if is_store:
            routing_key = "store"
            exchange_name = self._fallback_store_exchange_name

        assert isinstance(routing_key, str)
        routing_key = self._get_routing_key(routing_key, "retry.")

        updated_headers = self.send_message_to(
            exchange_name, ch, routing_key, properties, body
        )
        return ConsumerDerivedAction(
            action="send_to_retry",
            exchange_name=exchange_name,
            routing_key=routing_key,
            headers=updated_headers,
        )

    def send_to_dead_letter(
        self,
        ch: BlockingChannel,
        method: Basic.Deliver,
        properties: BasicProperties,
        body: bytes,
    ) -> ConsumerDerivedAction:
        self.printer.print_action("send_to_dead_letter")
        exchange_name = RabbitMqExchangeNameFormatter.dead_letter(self.exchange_name)
        assert isinstance(method.routing_key, str)
        routing_key = self._get_routing_key(method.routing_key, "dead_letter.")
        updated_headers = self.send_message_to(
            exchange_name, ch, routing_key, properties, body
        )
        return ConsumerDerivedAction(
            action="send_to_dead_letter",
            exchange_name=exchange_name,
            routing_key=routing_key,
            headers=updated_headers,
        )

    def send_message_to(
        self,
        exchange_name: str,
        ch: BlockingChannel,
        routing_key: str,
        properties: BasicProperties,
        body: bytes,
    ) -> Dict[str, Any]:
        if properties.headers:
            redelivery_count = properties.headers.get("redelivery_count", 0)
            properties.headers["redelivery_count"] = redelivery_count + 1
        else:
            properties.headers = {"redelivery_count": 1}

        self.printer.print_send_message_to(
            exchange_name, routing_key, dict(properties.headers)
        )

        ch.basic_publish(
            exchange=exchange_name,
            routing_key=routing_key,
            body=body,
            properties=properties,
        )

        return dict(properties.headers)

    def stop(self) -> None:
        """
        Stop the process of consuming messages from RabbitMQ.
        """

        def _log_stop_exception(e: Exception) -> None:
            message = f"Error stopping RabbitMQMessageConsumer: {repr(e.__class__)} {e} | {traceback.format_exc()}"
            logger.error(message)

        if self._thread and self._thread.is_alive():
            self._unsubscribe_all()
            try:
                self._thread.join()
                self._thread = None
            except Exception as e:
                _log_stop_exception(e)

    def _unsubscribe_all(self) -> None:
        def _stop_consuming_consumer_channel() -> None:
            try:
                self._channel.stop_consuming()
                self._channel.cancel()
            except ValueError:
                pass

        def _await_for_stop_consuming_consumer_channel() -> None:
            sleep(2.0)

        self.connector.get_connection(self.rabbitmq_key).call_later(
            0, _stop_consuming_consumer_channel
        )

        _await_for_stop_consuming_consumer_channel()

    def unsubscribe_subscriber_on_queue(self, queue_name: str) -> None:
        subscriber_item = self.subscribers.get(queue_name)
        if subscriber_item is None:
            raise IndexError(
                f"Cannot unsubscribe an nonexistent queue ({queue_name}). Please, check configured consumers ({list(self.subscribers.keys())})"
            )

        def _unsubscribe_handler_on_queue() -> None:
            assert isinstance(subscriber_item, SubscriberItem)
            self._channel.basic_cancel(consumer_tag=subscriber_item.consumer_tag)

        self._do_it_in_consumer_thread(_unsubscribe_handler_on_queue)

    def resume_subscriber_on_queue(self, queue_name: str) -> None:
        subscriber_item = self.subscribers.get(queue_name)
        if subscriber_item is None:
            raise IndexError(
                f"Cannot resume an nonexistent queue ({queue_name}). Please, check configured consumers ({list(self.subscribers.keys())})"
            )

        def _resume_handler_on_queue() -> None:
            assert isinstance(subscriber_item, SubscriberItem)
            subscriber_item.consumer_tag = self._channel.basic_consume(
                queue=subscriber_item.queue_name,
                on_message_callback=self.consumer(
                    subscriber_item.subscriber, subscriber_item.is_store
                ),
            )

        self._do_it_in_consumer_thread(_resume_handler_on_queue)

    def _do_it_in_consumer_thread(self, action: Callable[..., None]) -> None:
        def _execute_action() -> None:
            try:
                action()
            except ValueError:
                pass

        def _await_for_thread() -> None:
            sleep(2.0)

        self.connector.get_connection(self.rabbitmq_key).call_later(0, _execute_action)

        _await_for_thread()



from datetime import datetime
from typing import Any, cast

from petisco.base.misc.singleton import Singleton


class ApplicationInfo(metaclass=Singleton):
    name: str
    organization: str
    version: str
    deployed_at: datetime

    def __init__(self, **kwargs: Any) -> None:
        self.name = str(kwargs.get("name"))
        self.organization = str(kwargs.get("organization"))
        self.version = str(kwargs.get("version"))
        self.deployed_at = cast(datetime, kwargs.get("deployed_at"))



from abc import abstractmethod
from typing import Any

from meiga import AnyResult, NotImplementedMethodError

from petisco.base.misc.interface import Interface


class AppService(Interface):
    @abstractmethod
    def execute(self, *args: Any, **kwargs: Any) -> AnyResult:
        return NotImplementedMethodError



from typing import Any, Dict, Union

from petisco.extra.rabbitmq.shared.specific_queue_config import SpecificQueueConfig


class QueueConfig:
    @staticmethod
    def from_dict(
        specific_queue_configs_dict: Dict[str, Any],
        default_retry_ttl: int,
        default_main_ttl: int,
    ) -> "QueueConfig":
        specific_queue_configs = {}
        for key, config_dict in specific_queue_configs_dict.items():
            specific_queue_configs[key] = SpecificQueueConfig.from_dict(
                config_dict, default_retry_ttl, default_main_ttl
            )

        return QueueConfig(
            default_retry_ttl=default_retry_ttl,
            default_main_ttl=default_main_ttl,
            specific_queue_configs=specific_queue_configs,
        )

    @staticmethod
    def default(
        default_retry_ttl: int = 5000, default_main_ttl: int = 5000
    ) -> "QueueConfig":
        return QueueConfig(
            default_retry_ttl=default_retry_ttl,
            default_main_ttl=default_main_ttl,
            specific_queue_configs={},
        )

    def __init__(
        self,
        default_retry_ttl: int,
        default_main_ttl: int,
        specific_queue_configs: Dict[str, SpecificQueueConfig],
    ) -> None:
        self.default_retry_ttl = default_retry_ttl
        self.default_main_ttl = default_main_ttl
        self.specific_queue_configs = specific_queue_configs

    def info(self) -> Dict[str, Any]:
        return {
            name: specific_config.info()
            for name, specific_config in self.specific_queue_configs.items()
        }

    def get_retry_ttl(self, queue_name: str) -> Union[int, None]:
        for queue_config in self.specific_queue_configs.values():
            if queue_config.has_specific_config(queue_name):
                return queue_config.get_retry_ttl()
        return self.default_retry_ttl

    def get_main_ttl(self, queue_name: str) -> Union[int, None]:
        for queue_config in self.specific_queue_configs.values():
            if queue_config.has_specific_config(queue_name):
                return queue_config.get_main_ttl()
        return self.default_main_ttl



from typing import Any, Dict, Union


def create_text_meta(meta: Dict[str, Any]) -> Union[str, None]:
    text = None
    if meta:
        text = ""
        for key, value in meta.items():
            text += f"\n>{key}: {value}"
    return text



from petisco.extra.fastapi.is_fastapi_available import is_fastapi_available

__all__ = []

if is_fastapi_available():
    from petisco.extra.fastapi.application.fastapi_application import FastApiApplication
    from petisco.extra.fastapi.controller.fastapi_controller import FastAPIController
    from petisco.extra.fastapi.controller.fastapi_default_response import (
        FASTAPI_DEFAULT_RESPONSE,
    )
    from petisco.extra.fastapi.testing.assert_http_exception import (
        assert_http_exception,
    )

    __all__ = [
        "FastAPIController",
        "assert_http_exception",
        "FASTAPI_DEFAULT_RESPONSE",
        "FastApiApplication",
    ]



import json
from typing import Any, Dict, Optional, Type, Union

from petisco.base.domain.message.message import Message


class DomainEvent(Message):
    """
    A base class to model your domain events.
    Define your Domain Events to express what happened in your domain.
    """

    def __init__(self, **data: Any) -> None:
        super().__init__()
        self._set_attributes(**data)
        self.type = "domain_event"

    @staticmethod
    def from_dict(
        message_data: Dict[str, Any],
        target_type: Union[Type["DomainEvent"], None] = None,
    ) -> "DomainEvent":
        target_type = DomainEvent if target_type is None else target_type
        data = message_data.get("data")
        domain_event = target_type()
        domain_event._set_data(**data)  # type: ignore
        return domain_event

    @staticmethod
    def from_json(
        message_json: Union[str, bytes],
        target_type: Optional[Type["DomainEvent"]] = None,
    ) -> "DomainEvent":
        event_dict = json.loads(message_json)
        return DomainEvent.from_dict(event_dict, target_type)

    def __repr__(self) -> str:
        return self.to_str(class_name="DomainEvent", type="domain_event")



from abc import abstractmethod
from typing import Generic, List, TypeVar

from meiga import BoolResult, Error, NotImplementedMethodError, Result

from petisco.base.application.patterns.repository import Repository
from petisco.base.domain.model.aggregate_root import AggregateRoot
from petisco.base.domain.model.uuid import Uuid

AggregateRootType = TypeVar("AggregateRootType", bound=AggregateRoot)


class CrudRepository(Generic[AggregateRootType], Repository):
    @abstractmethod
    def save(self, aggregate_root: AggregateRootType) -> BoolResult:
        return NotImplementedMethodError

    @abstractmethod
    def retrieve(self, aggregate_id: Uuid) -> Result[AggregateRootType, Error]:
        return NotImplementedMethodError

    @abstractmethod
    def update(self, aggregate_root: AggregateRootType) -> BoolResult:
        return NotImplementedMethodError

    @abstractmethod
    def remove(self, aggregate_id: Uuid) -> BoolResult:
        return NotImplementedMethodError

    @abstractmethod
    def retrieve_all(self) -> Result[List[AggregateRootType], Error]:
        return NotImplementedMethodError



import os
from typing import Any, Dict, Union

from petisco.base.misc.builder import Builder


class Dependency:
    """
    Data structure to define how our dependency (Repository, AppService, etc) is built.

    Definition of default Builder is Mandatory, create alternative builder with envar_modifier is optional.
    """

    name: str
    default_builder: Builder[Any]
    envar_modifier: Union[str, None] = None
    builders: Union[Dict[str, Builder[Any]], None] = None

    def __init__(
        self,
        name: str,
        default_builder: Builder[Any],
        envar_modifier: Union[str, None] = None,
        builders: Union[Dict[str, Builder[Any]], None] = None,
    ):
        self.name = name
        self.default_builder = default_builder
        self.envar_modifier = envar_modifier
        self.builders = builders

    def get_instance(self) -> Any:
        """
        Returns an instance of the dependency.
        """
        if not self.envar_modifier:
            return self.default_builder.build()

        modifier = os.getenv(self.envar_modifier)
        if not modifier or not self.builders or modifier not in self.builders:
            return self.default_builder.build()
        else:
            builder = self.builders.get(modifier)
            assert isinstance(
                builder, Builder
            ), "Oh no! Dependency builder is corrupted!"
            instance = builder.build()
            return instance



from petisco.base.domain.errors.domain_error import DomainError


class CriticalError(DomainError):
    """
    A base class to define critical errors.

    If you inherit from CriticalError and uses NotifierMiddleware in your operations (Controllers or Subscribers), when
    you UseCase fails with these errors the system will notify the issue.
    """

    ...



import json
from typing import Any, Dict, Optional, Type, Union

from petisco.base.domain.message.message import Message


class Command(Message):
    """
    A base class to model your Command.
    An operation that effects some change to the system (for example, modify the persistence).
    An operation that intentionally creates a side effect.
    """

    def __init__(self, **data: Any) -> None:
        self._set_data()
        self._set_attributes(**data)
        self.type = "command"

    @staticmethod
    def from_dict(
        message_data: Dict[str, Any], target_type: Optional[Type["Command"]] = None
    ) -> "Command":
        target_type = Command if target_type is None else target_type
        data = message_data.get("data")
        command = target_type()
        command._set_data(**data)  # type: ignore
        return command

    @staticmethod
    def from_json(
        message_json: Union[str, bytes], target_type: Optional[Type["Command"]] = None
    ) -> "Command":
        event_dict = json.loads(message_json)
        return Command.from_dict(event_dict, target_type)

    def __repr__(self) -> str:
        return self.to_str(class_name="Command", type="domain_event")



class SqliteConnection:
    def __init__(self, server_name: str, database_name: str, url: str):
        self.server_name = server_name
        self.database_name = database_name
        self.url = url

    @staticmethod
    def create(server_name: str, database_name: str) -> "SqliteConnection":
        url = f"{server_name}:///{database_name}"
        return SqliteConnection(server_name, database_name, url)



from __future__ import annotations

import copy
from abc import ABC, abstractmethod
from typing import Any, Dict, Generic, TypeVar

from petisco.base.domain.message.message import Message

TypeMessage = TypeVar("TypeMessage", bound=Message)


class MessageBus(ABC, Generic[TypeMessage]):
    """
    A base class to implement an infrastructure-based bus to publish or dispatch messages.
    """

    @classmethod
    def __repr__(cls) -> str:
        return cls.__name__

    @classmethod
    def info(cls) -> dict[str, str]:
        return {"name": cls.__name__}

    def _set_additional_meta(self, meta: dict[str, Any]) -> None:
        self.additional_meta = meta

    def with_meta(self, meta: dict[str, Any]) -> MessageBus[Message]:
        if not isinstance(meta, Dict):
            raise TypeError("MessageBus.with_meta() expect a dict")
        message_bus = copy.copy(self)
        message_bus._set_additional_meta(meta)
        return message_bus  # type: ignore

    def get_configured_meta(self) -> dict[str, Any]:
        configured_meta: dict[str, Any] = {}
        if hasattr(self, "additional_meta"):
            configured_meta = {**configured_meta, **self.additional_meta}
        return configured_meta

    @abstractmethod
    def publish(self, message: TypeMessage) -> None:
        """
        Publish one message
        """
        raise NotImplementedError

    @abstractmethod
    def retry_publish_only_on_store_queue(self, message: TypeMessage) -> None:
        raise NotImplementedError

    @abstractmethod
    def close(self) -> None:
        raise NotImplementedError

    def _check_is_message(self, message: TypeMessage) -> None:
        if not message or not issubclass(message.__class__, Message):
            raise TypeError("MessageBus only publishes Message objects")

    def publish_list(self, messages: list[TypeMessage]) -> None:
        for message in messages:
            self.publish(message)

    def retry_publish_list_only_on_store_queue(
        self, messages: list[TypeMessage]
    ) -> None:
        for message in messages:
            self.retry_publish_only_on_store_queue(message)



from abc import ABC, abstractmethod
from types import FunctionType
from typing import Any, Dict, List, Optional, Tuple, cast

from meiga import AnyResult, NotImplementedMethodError

from petisco.base.application.controller.error_map import ErrorMap
from petisco.base.application.middleware.middleware import Middleware
from petisco.base.misc.result_mapper import ResultMapper, default_failure_handler
from petisco.base.misc.wrapper import wrapper


def get_mapper(bases: Tuple[Any], config: Optional[Dict[str, Any]]) -> ResultMapper:
    mapper = ResultMapper()
    if config:
        for base in bases:
            method = getattr(base, "get_config_mapper", None)
            if method:
                mapper = method(config)

    else:
        for base in bases:
            method = getattr(base, "get_default_mapper", None)
            if method:
                mapper = method()

    return mapper


class MetaController(type, ABC):
    middlewares: List[Middleware] = []

    def __new__(
        mcs, name: str, bases: Tuple[Any], namespace: Dict[str, Any]
    ) -> "MetaController":
        config = namespace.get("Config")

        mapper = get_mapper(bases, config)

        if "execute" not in namespace:
            raise NotImplementedError(
                "Petisco Controller must implement an execute method"
            )

        new_namespace = {}
        for attributeName, attribute in namespace.items():
            if isinstance(attribute, FunctionType) and attribute.__name__ == "execute":
                attribute = wrapper(attribute, name, config, mapper)
            new_namespace[attributeName] = attribute

        return super().__new__(mcs, name, bases, new_namespace)

    @abstractmethod
    def execute(self, *args: Any, **kwargs: Any) -> AnyResult:
        return NotImplementedMethodError


class Controller(metaclass=MetaController):
    """
    A base class for creating controllers.
    Inherit from this class to convert to domain the request values, configure middlewares and instantiate and execute
     a UseCase.
    """

    @staticmethod
    def get_default_mapper() -> ResultMapper:
        return ResultMapper()

    @staticmethod
    def get_config_mapper(config: Dict[str, Any]) -> ResultMapper:
        return ResultMapper(
            error_map=cast(ErrorMap, getattr(config, "error_map", None)),
            success_handler=getattr(config, "success_handler", lambda result: result),
            failure_handler=default_failure_handler,
        )

    @abstractmethod
    def execute(self, *args: Tuple[str, ...], **kwargs: Dict[str, Any]) -> AnyResult:
        return NotImplementedMethodError



from typing import Any, NoReturn, Type, TypeVar

from pydantic import validator
from pydantic.main import BaseModel

from petisco.base.domain.errors.defaults.invalid_value_object import InvalidValueObject

TypeValueObject = TypeVar("TypeValueObject", bound="ValueObject")


class ValueObject(BaseModel):
    """
    A base class to define ValueObject

    It is small object that represents a simple entity whose equality is not based on identity.
    """

    value: Any

    def __init__(self, value: Any, **data: Any) -> None:
        super().__init__(value=value, **data)

    def dict(self, **kwargs: Any) -> Any:
        return self.value

    def __setattr__(self, name: str, value: Any) -> NoReturn:
        raise TypeError("ValueObject objects are immutable")

    def __hash__(self) -> int:
        return hash(self.value)

    @classmethod
    def from_value(cls: Type[TypeValueObject], value: Any) -> TypeValueObject:
        return cls(value=value)

    @validator("value")
    def validate_value(cls, value: Any) -> Any:
        if value is None:
            raise InvalidValueObject()
        return value



# class SingletonMeta(type):
#     """
#     The Singleton class can be implemented in different ways in Python. Some
#     possible methods include: base class, decorator, metaclass. We will use the
#     metaclass because it is best suited for this purpose.
#     """
#
#     _instances = {}
#
#     def __call__(cls, *args: Any, **kwargs: Any):
#         """
#         Possible changes to the value of the `__init__` argument do not affect
#         the returned instance.
#         """
#         if cls not in cls._instances:
#             instance = super().__call__(*args: Any, **kwargs: Any)
#             cls._instances[cls] = instance
#         return cls._instances[cls]
#
#
# class Singleton(metaclass=SingletonMeta):
#     pass
from typing import Any, Dict


class Singleton(type):
    _instances: Dict[Any, Any] = {}

    def __call__(cls, *args: Any, **kwargs: Any) -> Any:
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]

    def clear(cls) -> None:
        try:
            del Singleton._instances[cls]
        except KeyError:
            pass



from abc import abstractmethod
from typing import Any

from petisco.base.misc.interface import Interface


class MessageChaos(Interface):
    @abstractmethod
    def nack_simulation(self, *args: Any, **kwargs: Any) -> bool:
        raise NotImplementedError

    @abstractmethod
    def failure_simulation(self, *args: Any, **kwargs: Any) -> bool:
        raise NotImplementedError

    @abstractmethod
    def delay(self) -> None:
        raise NotImplementedError



from abc import abstractmethod

from petisco.base.misc.interface import Interface


class ApplicationConfigurer(Interface):
    def __init__(self, execute_after_dependencies: bool = False) -> None:
        self.execute_after_dependencies = execute_after_dependencies

    @abstractmethod
    def execute(self, testing: bool = False) -> None:
        raise NotImplementedError()



from abc import abstractmethod
from typing import Any, Dict

from meiga import AnyResult, NotImplementedMethodError

from petisco.base.application.controller.controller import Controller
from petisco.base.misc.result_mapper import ResultMapper
from petisco.extra.fastapi.controller.fastapi_result_mapper import FastAPIResultMapper


class FastAPIController(Controller):
    @staticmethod
    def get_default_mapper() -> ResultMapper:
        return FastAPIResultMapper.default()

    @staticmethod
    def get_config_mapper(config: Dict[str, Any]) -> ResultMapper:
        return FastAPIResultMapper.from_config(config)

    @abstractmethod
    def execute(self, *args: Any, **kwargs: Any) -> AnyResult:
        return NotImplementedMethodError



from petisco.base.domain.message.message import Message


class RabbitMqMessageQueueNameFormatter:
    @staticmethod
    def format(message: Message, exchange_name: str = None) -> str:
        message_name = message.name.replace(".", "_")
        message_type = message.type if message.type != "domain_event" else "event"
        message_format = f"{message.version}.{message_type}.{message_name}"
        return f"{exchange_name}.{message_format}" if exchange_name else message_name

    @staticmethod
    def format_retry(message: Message, exchange_name: str = None) -> str:
        queue_name = RabbitMqMessageQueueNameFormatter.format(message, exchange_name)
        return f"retry.{queue_name}"

    @staticmethod
    def format_dead_letter(message: Message, exchange_name: str = None) -> str:
        queue_name = RabbitMqMessageQueueNameFormatter.format(message, exchange_name)
        return f"dead_letter.{queue_name}"



import sys
import traceback
from typing import Any, Dict, List, Optional, Union

from meiga import Error


class UnknownError(Error):
    """
    A base class to define unknown errors.

    If UnknownError is caught and uses NotifierMiddleware in your operations (Controllers or Subscribers), when
    you UseCase fails the system will notify the issue.
    """

    def __init__(
        self,
        exception: Exception,
        input_parameters: Any = None,
        executor: Any = None,
        traceback: Any = None,
        filename: Any = None,
        lineno: Any = None,
        filter_parameters: Union[List[str], None] = None,
        meta: Union[Dict[str, Any], None] = None,
    ) -> None:
        self.message = f"{exception.__class__.__name__}: {str(exception)}"
        self.input_parameters = self._sanitize_input_params(input_parameters)
        self._filter_input_parameters(filter_parameters)
        self.exception = exception
        self.executor = executor
        self.traceback = traceback
        self.filename = filename
        self.lineno = lineno
        self.meta = meta if meta else {}

    def _sanitize_input_params(self, input_parameters: Any) -> Any:
        if isinstance(input_parameters, tuple):
            return {
                f"param_{i + 1}": param if not isinstance(param, bytes) else "bytes"
                for i, param in enumerate(input_parameters)
            }
        elif isinstance(input_parameters, dict):
            return {
                k: v if not isinstance(v, bytes) else "bytes"
                for k, v in input_parameters.items()
            }
        else:
            return None

    def _filter_input_parameters(
        self, filter_parameters: Union[List[str], None] = None
    ) -> None:
        if filter_parameters:
            self.input_parameters = {
                k: v
                for k, v in self.input_parameters.items()
                if k not in filter_parameters
            }

    def __repr__(self) -> str:
        executor_str = f" ({self.executor})" if self.executor else ""
        traceback_str = f"\n{self.traceback}" if self.traceback else ""
        input_parameters_str = (
            f"\nInput Parameters: {str(self.input_parameters)}"
            if self.input_parameters
            else ""
        )
        filename_str = f"\n{self.filename}" if self.filename else ""
        lineno_str = f"\n{self.lineno}" if self.lineno else ""
        return f"{self.__class__.__name__}{executor_str}: {self.message}.{traceback_str}.{input_parameters_str}{filename_str}{lineno_str}"

    @classmethod
    def from_exception(
        cls,
        exception: Exception,
        arguments: Any,
        class_name: Optional[str] = None,
    ) -> "UnknownError":
        _, _, tb = sys.exc_info()
        tb = traceback.extract_tb(tb)[-1]  # type: ignore
        filename = tb.filename if tb and hasattr(tb, "filename") else None  # type: ignore
        lineno = tb.lineno if tb and hasattr(tb, "lineno") else None  # type: ignore

        unknown_error = cls(
            exception=exception,
            input_parameters=arguments,
            executor=class_name,
            traceback=traceback.format_exc(),
            filename=filename,
            lineno=lineno,
        )
        return unknown_error



from typing import Dict, Optional

from meiga import Error


class DomainError(Error):
    """
    A base class to define Domain Errors
    """

    def __init__(
        self,
        uuid_value: Optional[str] = None,
        additional_info: Optional[Dict[str, str]] = None,
    ):
        self.uuid_value = uuid_value
        self.additional_info = additional_info
        self._additional_detail = ""
        self._specific_detail = "DomainError"
        self._set_detail()

    def _set_detail(self) -> None:
        if self.uuid_value:
            self._additional_detail += f" ({self.uuid_value})"

        if self.additional_info:
            self._additional_detail += f" [{self.additional_info}]"

    @classmethod
    def get_specify_detail(cls) -> str:
        return cls.__name__

    def detail(self) -> str:
        return f"{self.get_specify_detail()}{self._additional_detail}"

    def __str__(self) -> str:
        return f"{self.get_specify_detail()}{self._additional_detail}"

    def __repr__(self) -> str:
        return f"{self.get_specify_detail()}{self._additional_detail}"



import re

from petisco.base.domain.message.message_subscriber_info import MessageSubscriberInfo


class RabbitMqMessageSubscriberQueueNameFormatter:
    @staticmethod
    def format(
        subscriber_info: MessageSubscriberInfo, exchange_name: str = None
    ) -> str:
        message_name = (
            re.sub(r"(?<!^)(?=[A-Z])", "_", subscriber_info.message_name)
            .lower()
            .replace(".", "_")
        )
        message_type = (
            subscriber_info.message_type
            if subscriber_info.message_type != "domain_event"
            else "event"
        )
        message_format = (
            f"{subscriber_info.message_version}.{message_type}.{message_name}"
        )
        return f"{exchange_name}.{message_format}" if exchange_name else message_format

    @staticmethod
    def format_retry(
        subscriber_info: MessageSubscriberInfo, exchange_name: str = None
    ) -> str:
        queue_name = RabbitMqMessageSubscriberQueueNameFormatter.format(
            subscriber_info, exchange_name
        )
        return f"retry.{queue_name}"

    @staticmethod
    def format_dead_letter(
        subscriber_info: MessageSubscriberInfo, exchange_name: str = None
    ) -> str:
        queue_name = RabbitMqMessageSubscriberQueueNameFormatter.format(
            subscriber_info, exchange_name
        )
        return f"dead_letter.{queue_name}"



from meiga import AnyResult

from petisco.base.application.middleware.middleware import Middleware


class PrintMiddleware(Middleware):
    def before(self) -> None:
        print(
            f"{self.wrapped_class_name} -> Start | Params {dict(self.wrapped_class_input_arguments)}"
        )

    def after(self, result: AnyResult) -> None:
        print(f"{self.wrapped_class_name} -> End | {result}")



from typing import Any, Dict, Union

from petisco.extra.rabbitmq.shared.rabbitmq_connector import RabbitMqConnector


class RabbitMqDeclarer:
    def __init__(self, connector: RabbitMqConnector, channel_name: str):
        self._connector = connector
        self._channel_name = channel_name

    def declare_exchange(
        self, exchange_name: str, exchange_type: str = "topic"
    ) -> None:
        channel = self._connector.get_channel(self._channel_name)
        try:
            channel.exchange_declare(
                exchange=exchange_name, exchange_type=exchange_type, durable=True
            )
        except Exception as error:
            raise TypeError(
                f"RabbitMQEventPublisher: Cannot create the exchange ({exchange_name})\n{error}"
            )
        channel.close()

    def delete_exchange(self, exchange_name: str) -> None:
        channel = self._connector.get_channel(self._channel_name)
        channel.exchange_delete(exchange_name)
        channel.close()

    def declare_queue(
        self,
        queue_name: str,
        dead_letter_exchange: Union[str, None] = None,
        dead_letter_routing_key: Union[str, None] = None,
        message_ttl: Union[int, None] = None,
    ) -> str:
        channel = self._connector.get_channel(self._channel_name)

        queue_arguments: Dict[str, Any] = {}
        if dead_letter_exchange:
            queue_arguments["x-dead-letter-exchange"] = dead_letter_exchange

        if dead_letter_routing_key:
            queue_arguments["x-dead-letter-routing-key"] = dead_letter_routing_key

        if message_ttl:
            queue_arguments["x-message-ttl"] = message_ttl

        result = channel.queue_declare(
            queue=queue_name, arguments=queue_arguments, durable=True
        )
        channel.close()

        queue_name: str = result.method.queue

        return queue_name

    def bind_queue(self, exchange_name: str, queue_name: str, routing_key: str) -> None:
        channel = self._connector.get_channel(self._channel_name)

        channel.queue_bind(
            exchange=exchange_name, queue=queue_name, routing_key=routing_key
        )
        channel.close()

    def delete_queue(self, queue_name: str) -> None:
        channel = self._connector.get_channel(self._channel_name)
        channel.queue_delete(queue_name)
        channel.close()



from abc import ABC
from copy import copy
from typing import List, Union

from pydantic import BaseModel, Field, NonNegativeInt, PrivateAttr, validator

from petisco.base.domain.message.domain_event import DomainEvent
from petisco.base.domain.model.uuid import Uuid

DEFAULT_VERSION = 1


class AggregateRoot(ABC, BaseModel):
    """
    A base class to define AggregateRoot

    It is a cluster of associated entities that are treated as a unit for the purpose of data changes.
    """

    aggregate_id: Uuid = Field(default=Uuid.v4())
    aggregate_version: NonNegativeInt = Field(default=DEFAULT_VERSION)
    _domain_events: List[DomainEvent] = PrivateAttr(default=[])

    @validator("aggregate_id", pre=True, always=True)
    def set_aggregate_id(cls, v: Union[str, Uuid]) -> Uuid:
        v = Uuid(v) if isinstance(v, str) else v
        return v or Uuid.v4()

    @validator("aggregate_version", pre=True, always=True)
    def set_aggregate_version(cls, v: NonNegativeInt) -> NonNegativeInt:
        return v or DEFAULT_VERSION

    def record(self, domain_event: DomainEvent) -> None:
        """
        Record something that happened is our domain related with the aggregate (a DomainEvent).
        """
        self._domain_events.append(domain_event)

    def clear_domain_events(self) -> None:
        """
        Clear all domain events recorded.
        """
        self._domain_events = []

    def pull_domain_events(self) -> List[DomainEvent]:
        """
        Returns and clear all domain events recorded.
        """
        domain_events = copy(self._domain_events)
        self.clear_domain_events()
        return domain_events

    def get_domain_events(self) -> List[DomainEvent]:
        """
        Returns all domain events recorded.
        """
        return self._domain_events

    def get_first_domain_event(self) -> Union[DomainEvent, None]:
        """
        Returns only the first domain event recorded.
        """
        return self._domain_events[0] if len(self._domain_events) > 0 else None

    def get_last_domain_event(self) -> Union[DomainEvent, None]:
        """
        Returns only the last domain event recorded.
        """
        return self._domain_events[-1] if len(self._domain_events) > 0 else None



from typing import Any, Dict

from petisco.base.domain.message.chaos.message_chaos import MessageChaos


class NotImplementedMessageChaos(MessageChaos):
    def info(self) -> Dict[str, Any]:
        return {
            "EVENT_CHAOS_PERCENTAGE_SIMULATE_NACK": "NOT_IMPLEMENTED",
            "EVENT_CHAOS_DELAY_AFTER_OPERATION_SECONDS": "NOT_IMPLEMENTED",
            "EVENT_CHAOS_PERCENTAGE_SIMULATE_FAILURES": "NOT_IMPLEMENTED",
        }

    def nack_simulation(self, *args: Any, **kwargs: Any) -> bool:
        return False

    def failure_simulation(self, *args: Any, **kwargs: Any) -> bool:
        return False

    def delay(self) -> None:
        pass



from abc import abstractmethod
from typing import List, Type

from meiga import BoolResult

from petisco.base.domain.message.domain_event import DomainEvent
from petisco.base.domain.message.message_subscriber import MessageSubscriber


class DomainEventSubscriber(MessageSubscriber):
    """
    A base class to model your events subscribers.
    Inherit from this class to parser the domain event, configure middlewares and instantiate and execute a UseCase.
    """

    @abstractmethod
    def subscribed_to(self) -> List[Type[DomainEvent]]:
        """
        returns the list of events that we want to subscribe
        """
        raise NotImplementedError()

    @abstractmethod
    def handle(self, domain_event: DomainEvent) -> BoolResult:
        """
        returns True if the event was processed or False if it could not be processed
        """
        raise NotImplementedError()



from petisco.extra.elastic.is_elastic_available import is_elastic_available
from petisco.extra.elastic_apm.is_elastic_apm_available import is_elastic_apm_available

elastic_apm = list()
if is_elastic_available() or is_elastic_apm_available():
    from petisco.extra.elastic_apm.elastic_apm_monitoring_app_service import (
        ElasticApmMonitoringAppService,
    )

    elastic_apm = ["ElasticApmMonitoringAppService"]
else:
    elastic_apm = []

__all__ = elastic_apm



from abc import ABC, abstractmethod
from functools import wraps
from types import FunctionType
from typing import Any, Callable, Dict, Tuple

import elasticapm
from meiga import AnyResult, Error, Failure, NotImplementedMethodError
from meiga.on_failure_exception import OnFailureException

from petisco.base.application.use_case.use_case_uncontrolled_error import (
    UseCaseUncontrolledError,
)


def wrapper(
    method: Callable[..., AnyResult], wrapped_class_name: str
) -> Callable[..., AnyResult]:
    @wraps(method)
    def wrapped(*args: Any, **kwargs: Any) -> AnyResult:
        try:
            return method(*args, **kwargs)
        except OnFailureException as exception:
            # TODO create issue to add typing on OnFailureException meiga class
            return exception.result  # type: ignore
        except Error as error:
            return Failure(error)
        except Exception as exception:
            uncontrolled_error = UseCaseUncontrolledError.from_exception(
                exception=exception, arguments=args, class_name=wrapped_class_name
            )
            client = elasticapm.get_client()
            if client:
                client.capture_exception()

            return Failure(uncontrolled_error)

    return wrapped


class MetaUseCase(type, ABC):
    def __new__(
        mcs, name: str, bases: Tuple[Any], namespace: Dict[str, Any]
    ) -> "MetaUseCase":
        new_class_dict = {}
        if "execute" not in namespace:
            raise NotImplementedError(
                "Petisco UseCase must implement an execute method"
            )

        for attributeName, attribute in namespace.items():
            if isinstance(attribute, FunctionType) and attribute.__name__ == "execute":
                attribute = wrapper(attribute, name)
            new_class_dict[attributeName] = attribute
        return type.__new__(mcs, name, bases, new_class_dict)

    @abstractmethod
    def execute(self, *args: Any, **kwargs: Any) -> AnyResult:
        return NotImplementedMethodError


class UseCase(metaclass=MetaUseCase):
    @abstractmethod
    def execute(self, *args: Any, **kwargs: Any) -> AnyResult:
        return NotImplementedMethodError



from contextlib import contextmanager
from typing import Callable

from loguru import logger


def elastic_session_scope_provider(session) -> Callable:
    from elasticsearch import RequestError

    @contextmanager
    def session_scope():
        try:
            yield session
        except ConnectionRefusedError as e:
            logger.error(e)
            raise e
        except RequestError as e:
            logger.error(e)
            raise e
        except Exception as e:
            logger.error(e)
            raise e

    return session_scope



from uuid import uuid4

import validators
from pydantic import validator

from petisco.base.domain.errors.defaults.invalid_uuid import InvalidUuid
from petisco.base.domain.model.value_object import ValueObject


class Uuid(ValueObject):
    """
    A base class to define Uuid

    Use it to identify domain entities
    """

    value: str

    @validator("value")
    def validate_value(cls, value: str) -> str:
        if value is None or not validators.uuid(value):
            raise InvalidUuid(uuid_value=value)
        return value

    @classmethod
    def v4(cls) -> "Uuid":
        return cls(value=str(uuid4()))



from petisco.extra.sqlmodel.is_sqlmodel_available import is_sqlmodel_available

if is_sqlmodel_available():
    from petisco.extra.sqlmodel.sqlmodel_crud_repository import SQLModelCrudRepository

    __all__ = ["SQLModelCrudRepository"]
else:
    __all__ = []



from abc import abstractmethod
from typing import List, Type

from meiga import BoolResult

from petisco.base.domain.message.command import Command
from petisco.base.domain.message.message_subscriber import MessageSubscriber

# Unlike in the domain event subscriber,
# in the command handler the command -> command handler
# relationship is 1 to 1 and that is why we use Type[Command] instead of List[Type[Command]]
# and overwrite get_message_subscribers_info base method
from petisco.base.domain.message.message_subscriber_info import MessageSubscriberInfo


class CommandSubscriber(MessageSubscriber):
    """
    A base class to model your command subscribers (or also called CommandHandlers).
    Inherit from this class to parser the command, configure middlewares and instantiate and execute a UseCase.
    """

    @abstractmethod
    def subscribed_to(self) -> Type[Command]:
        raise NotImplementedError()

    @abstractmethod
    def handle(self, command: Command) -> BoolResult:  # type: ignore
        raise NotImplementedError()

    def get_message_subscribers_info(self) -> List[MessageSubscriberInfo]:
        return [MessageSubscriberInfo.from_class_type(self.subscribed_to())]



from abc import abstractmethod

from petisco.base.domain.message.command import Command
from petisco.base.domain.message.message_bus import MessageBus


class CommandBus(MessageBus[Command]):
    """
    A base class to implement an infrastructure-based bus to dispatch commands.
    """

    @abstractmethod
    def dispatch(self, command: Command) -> None:
        """
        Dispatch one Command
        """
        raise NotImplementedError

    def publish(self, command: Command) -> None:
        self.dispatch(command)

    def retry_publish_only_on_store_queue(self, command: Command) -> None:
        pass

    def _check_is_command(self, command: Command) -> None:
        if not command or not issubclass(command.__class__, Command):
            raise TypeError("CommandBus only publishes DomainEvent objects")

    @abstractmethod
    def close(self) -> None:
        raise NotImplementedError



from fnmatch import fnmatch
from typing import Any, Dict, Union


class SpecificQueueConfig:
    @staticmethod
    def from_dict(
        kdict: Dict[str, Any], default_retry_ttl: int, default_main_ttl: int
    ) -> "SpecificQueueConfig":
        return SpecificQueueConfig(
            wildcard=kdict.get("when_queue_contains", None),
            specific_retry_ttl=kdict.get("use", {}).get("retry_ttl", default_retry_ttl),
            specific_main_ttl=kdict.get("use", {}).get("main_ttl", default_main_ttl),
        )

    def __init__(
        self,
        wildcard: str,
        specific_retry_ttl: Union[int, None] = None,
        specific_main_ttl: Union[int, None] = None,
    ) -> None:
        self.wildcard = wildcard
        self.specific_retry_ttl = specific_retry_ttl
        self.specific_main_ttl = specific_main_ttl

    def info(self) -> Dict[str, Any]:
        return {
            "when_queue_contains": self.wildcard,
            "specific_retry_ttl": self.specific_retry_ttl,
            "specific_main_ttl": self.specific_main_ttl,
        }

    def has_specific_config(self, queue_name: str) -> bool:
        return fnmatch(queue_name, self.wildcard)

    def get_retry_ttl(self) -> Union[int, None]:
        return self.specific_retry_ttl

    def get_main_ttl(self) -> Union[int, None]:
        return self.specific_main_ttl
